"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Sandbox = exports.SandboxSettings = void 0;
const tslib_1 = require("tslib");
const analytics_next_1 = require("@segment/analytics-next");
const logger_1 = require("../../lib/logger");
const workerboxjs_1 = tslib_1.__importDefault(require("workerboxjs"));
const signals_runtime_1 = require("./signals-runtime");
const replace_base_url_1 = require("../../lib/replace-base-url");
/**
 * Proxy around the analytics client
 */
class AnalyticsRuntime {
    constructor() {
        this.calls = {
            page: [],
            identify: [],
            track: [],
            alias: [],
            screen: [],
            group: [],
        };
        // these methods need to be bound to the instance, rather than the prototype, in order to serialize correctly in the sandbox.
        this.track = (...args) => {
            // @ts-ignore
            const [eventName, props, options, cb] = (0, analytics_next_1.resolveArguments)(...args);
            this.calls.track.push([eventName, props, this.stamp(options), cb]);
        };
        this.identify = (...args) => {
            // @ts-ignore
            const [id, traits, options, cb] = (0, analytics_next_1.resolveUserArguments)(...args);
            this.stamp(options);
            this.calls.identify.push([id, traits, this.stamp(options), cb]);
        };
        this.alias = (...args) => {
            const [userId, previousId, options, cb] = (0, analytics_next_1.resolveAliasArguments)(
            // @ts-ignore
            ...args);
            this.calls.alias.push([userId, previousId, this.stamp(options), cb]);
        };
        this.group = (...args) => {
            // @ts-ignore
            const [id, traits, options, cb] = (0, analytics_next_1.resolveUserArguments)(...args);
            this.calls.group.push([id, traits, this.stamp(options), cb]);
        };
        this.page = (...args) => {
            const [category, name, props, options, cb] = (0, analytics_next_1.resolvePageArguments)(...args);
            this.stamp(options);
            this.calls.page.push([category, name, props, this.stamp(options), cb]);
        };
        this.screen = (...args) => {
            const [category, name, props, options, cb] = (0, analytics_next_1.resolvePageArguments)(...args);
            this.stamp(options);
            this.calls.screen.push([category, name, props, this.stamp(options), cb]);
        };
    }
    getCalls() {
        return this.calls;
    }
    /**
     * Stamp the context with the event origin to prevent infinite signal-event loops.
     */
    stamp(options) {
        if (!options) {
            options = {};
        }
        options.context = { ...options.context, __eventOrigin: { type: 'Signal' } };
        return options;
    }
}
class JavascriptSandbox {
    constructor() {
        this.workerbox = (0, workerboxjs_1.default)('');
    }
    async run(fn, scope) {
        try {
            const wb = await this.workerbox;
            await wb.run(fn, scope);
        }
        catch (err) {
            console.error('processSignal() error in sandbox', err, {
                fn,
            });
        }
    }
    async destroy() {
        const wb = await this.workerbox;
        await wb.destroy();
    }
}
class SandboxSettings {
    constructor(settings) {
        const edgeFnDownloadURLNormalized = settings.functionHost && settings.edgeFnDownloadURL
            ? (0, replace_base_url_1.replaceBaseUrl)(settings.edgeFnDownloadURL, `https://${settings.functionHost}`)
            : settings.edgeFnDownloadURL;
        if (!edgeFnDownloadURLNormalized && !settings.processSignal) {
            throw new Error('edgeFnDownloadURL or processSignal must be defined');
        }
        const fetch = settings.edgeFnFetchClient ?? globalThis.fetch;
        const processSignalNormalized = settings.processSignal
            ? Promise.resolve(settings.processSignal).then((str) => `globalThis.processSignal = ${str}`)
            : fetch(edgeFnDownloadURLNormalized).then((res) => res.text());
        this.processSignal = processSignalNormalized;
    }
}
exports.SandboxSettings = SandboxSettings;
class Sandbox {
    constructor(settings) {
        this.settings = settings;
        this.jsSandbox = new JavascriptSandbox();
    }
    async process(signal, signals) {
        const analytics = new AnalyticsRuntime();
        const scope = {
            analytics,
        };
        logger_1.logger.debug('processing signal', { signal, scope, signals });
        const code = [
            await this.settings.processSignal,
            `const createSignalsRuntime = ${signals_runtime_1.createSignalsRuntime.toString()}`,
            `const signals = createSignalsRuntime(${JSON.stringify(signals)})`,
            'try { processSignal(' +
                JSON.stringify(signal) +
                ', { analytics, signals }); } catch(err) { console.error("Process signal failed.", err); }',
        ].join('\n');
        await this.jsSandbox.run(code, scope);
        const calls = analytics.getCalls();
        logger_1.logger.debug('analytics calls', calls);
        return calls;
    }
}
exports.Sandbox = Sandbox;
//# sourceMappingURL=sandbox.js.map